# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.common import locking
from vmware_nsx.db import db as nsx_db
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_v3.implementation import lb_utils
from vmware_nsxlib.v3 import exceptions as nsxlib_exc

LOG = logging.getLogger(__name__)


class EdgeMemberManagerFromDict(base_mgr.Nsxv3LoadbalancerBaseManager):
    def _get_info_from_fip(self, context, fip):
        filters = {'floating_ip_address': [fip]}
        floating_ips = self.core_plugin.get_floatingips(context,
                                                        filters=filters)
        if floating_ips:
            return (floating_ips[0]['fixed_ip_address'],
                    floating_ips[0]['router_id'])
        else:
            msg = (_('Member IP %(fip)s is an external IP, and is expected to '
                     'be a floating IP') % {'fip': fip})
            raise n_exc.BadRequest(resource='lbaas-vip', msg=msg)

    def _get_updated_pool_members(self, context, lb_pool, member):
        network = lb_utils.get_network_from_subnet(
            context, self.core_plugin, member['subnet_id'])
        if network.get('router:external'):
            fixed_ip, router_id = self._get_info_from_fip(
                context, member['address'])
        else:
            fixed_ip = member['address']
        for m in lb_pool['members']:
            if m['ip_address'] == fixed_ip:
                m['display_name'] = member['name'][:219] + '_' + member['id']
                m['weight'] = member['weight']
                m['backup_member'] = member.get('backup', False)
        return lb_pool['members']

    def create(self, context, member, completor):
        with locking.LockManager.get_lock(
            'member-%s' % str(member['pool']['loadbalancer_id'])):
            self._member_create(context, member, completor)

    def _member_create(self, context, member, completor):
        lb_id = member['pool']['loadbalancer_id']
        pool_id = member['pool']['id']
        loadbalancer = member['pool']['loadbalancer']
        if not lb_utils.validate_lb_member_subnet(context, self.core_plugin,
                                                  member['subnet_id'],
                                                  loadbalancer):
            completor(success=False)
            msg = (_('Cannot add member %(member)s to pool as member subnet '
                     '%(subnet)s is neither public nor connected to the LB '
                     'router') %
                   {'member': member['id'], 'subnet': member['subnet_id']})
            raise n_exc.BadRequest(resource='lbaas-subnet', msg=msg)

        pool_client = self.core_plugin.nsxlib.load_balancer.pool
        service_client = self.core_plugin.nsxlib.load_balancer.service

        network = lb_utils.get_network_from_subnet(
            context, self.core_plugin, member['subnet_id'])
        if network.get('router:external'):
            fixed_ip, router_id = self._get_info_from_fip(
                context, member['address'])
            if not router_id:
                completor(success=False)
                msg = (_('Floating ip %(fip)s has no router') % {
                    'fip': member['address']})
                raise n_exc.BadRequest(resource='lbaas-vip', msg=msg)
        else:
            router_id = lb_utils.get_router_from_network(
                context, self.core_plugin, member['subnet_id'])
            fixed_ip = member['address']

        binding = nsx_db.get_nsx_lbaas_pool_binding(context.session,
                                                    lb_id, pool_id)
        if binding:
            lb_pool_id = binding.get('lb_pool_id')
            lb_binding = nsx_db.get_nsx_lbaas_loadbalancer_binding(
                context.session, lb_id)
            if not lb_binding:
                completor(success=False)
                msg = (_('Failed to get LB binding for member %s') %
                       member['id'])
                raise nsx_exc.NsxPluginException(err_msg=msg)
            if lb_binding.lb_router_id == lb_utils.NO_ROUTER_ID:
                # Need to attach the LB service to the router now
                # This will happen here in case of external vip
                nsx_router_id = nsx_db.get_nsx_router_id(context.session,
                                                         router_id)
                try:
                    # Make sure the NSX service router exists
                    if not self.core_plugin.verify_sr_at_backend(
                            context, router_id):
                        self.core_plugin.create_service_router(
                            context, router_id)

                    tags = lb_utils.get_tags(self.core_plugin, router_id,
                                             lb_const.LR_ROUTER_TYPE,
                                             member['tenant_id'],
                                             context.project_name)
                    service_client.update_service_with_attachment(
                        lb_binding.lb_service_id, nsx_router_id, tags=tags)
                    # TODO(asarfaty): Also update the tags
                except nsxlib_exc.ManagerError as e:
                    # This will happen if there is another service already
                    # attached to this router.
                    # This is currently a limitation.
                    completor(success=False)
                    msg = (_('Failed to attach router %(rtr)s to LB service '
                             '%(srv)s: %(e)s') %
                           {'rtr': router_id, 'srv': lb_binding.lb_service_id,
                            'e': e})
                    raise nsx_exc.NsxPluginException(err_msg=msg)
                # Update the nsx router in the DB binding
                nsx_db.update_nsx_lbaas_loadbalancer_binding(
                    context.session, lb_id, nsx_router_id)
                # Add rule to advertise external vips
                router = self.core_plugin.get_router(context, router_id)
                lb_utils.update_router_lb_vip_advertisement(
                    context, self.core_plugin, router, nsx_router_id)

            with locking.LockManager.get_lock('pool-member-%s' % lb_pool_id):
                lb_pool = pool_client.get(lb_pool_id)
                old_m = lb_pool.get('members', None)
                new_m = [{
                    'display_name': member['name'][:219] + '_' + member['id'],
                    'ip_address': fixed_ip,
                    'port': member['protocol_port'],
                    'weight': member['weight'],
                    'backup_member': member.get('backup', False)}]
                members = (old_m + new_m) if old_m else new_m
                pool_client.update_pool_with_members(lb_pool_id, members)

        else:
            completor(success=False)
            msg = (_('Failed to get pool binding to add member %s') %
                   member['id'])
            raise nsx_exc.NsxPluginException(err_msg=msg)

        completor(success=True)

    def update(self, context, old_member, new_member, completor):
        lb_id = old_member['pool']['loadbalancer_id']
        pool_id = old_member['pool']['id']
        pool_client = self.core_plugin.nsxlib.load_balancer.pool
        pool_binding = nsx_db.get_nsx_lbaas_pool_binding(
            context.session, lb_id, pool_id)
        if pool_binding:
            lb_pool_id = pool_binding.get('lb_pool_id')
            try:
                with locking.LockManager.get_lock('pool-member-%s' %
                                                  lb_pool_id):
                    lb_pool = pool_client.get(lb_pool_id)
                    updated_members = self._get_updated_pool_members(
                        context, lb_pool, new_member)
                    pool_client.update_pool_with_members(lb_pool_id,
                                                         updated_members)
            except Exception as e:
                with excutils.save_and_reraise_exception():
                    completor(success=False)
                    LOG.error('Failed to update member %(member)s: '
                              '%(err)s',
                              {'member': old_member['id'], 'err': e})
        completor(success=True)

    def delete(self, context, member, completor):
        lb_id = member['pool']['loadbalancer_id']
        pool_id = member['pool']['id']
        pool_client = self.core_plugin.nsxlib.load_balancer.pool
        pool_binding = nsx_db.get_nsx_lbaas_pool_binding(
            context.session, lb_id, pool_id)
        if pool_binding:
            lb_pool_id = pool_binding.get('lb_pool_id')
            try:
                with locking.LockManager.get_lock('pool-member-%s' %
                                                  lb_pool_id):
                    lb_pool = pool_client.get(lb_pool_id)
                    network = lb_utils.get_network_from_subnet(
                        context, self.core_plugin, member['subnet_id'])
                    if network.get('router:external'):
                        fixed_ip, router_id = self._get_info_from_fip(
                            context, member['address'])
                    else:
                        fixed_ip = member['address']
                    if 'members' in lb_pool:
                        m_list = lb_pool['members']
                        members = [m for m in m_list
                                if m['ip_address'] != fixed_ip]
                        pool_client.update_pool_with_members(lb_pool_id,
                                                            members)
            except nsxlib_exc.ResourceNotFound:
                pass
            except nsxlib_exc.ManagerError:
                completor(success=False)
                msg = _('Failed to remove member from pool on NSX backend')
                raise n_exc.BadRequest(resource='lbaas-member', msg=msg)

        completor(success=True)

    def delete_cascade(self, context, member, completor):
        # No action should be taken on members delete cascade
        pass



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import smtplib

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from oslo_config import cfg
from oslo_log import log
import stevedore

from neutron_lib import exceptions as n_exc
from vmware_nsx.common import locking

LOG = log.getLogger(__name__)
ALL_DUMMY_JOB_NAME = 'all'
ALL_DUMMY_JOB = {
    'name': ALL_DUMMY_JOB_NAME,
    'description': 'Execute all housekeepers',
    'enabled': True,
    'error_count': 0,
    'fixed_count': 0,
    'error_info': None}


class NsxHousekeeper(stevedore.named.NamedExtensionManager):
    def __init__(self, hk_ns, hk_jobs, hk_readonly, hk_readonly_jobs):
        self.global_readonly = hk_readonly
        self.readonly_jobs = hk_readonly_jobs
        self.email_notifier = None
        if (cfg.CONF.smtp_gateway and
                cfg.CONF.smtp_from_addr and
                cfg.CONF.snmp_to_list):
            self.email_notifier = HousekeeperEmailNotifier()

        self.results = {}

        if self.global_readonly:
            LOG.info('Housekeeper initialized in readonly mode')
        else:
            LOG.info('Housekeeper initialized')

        self.results = {}
        self.jobs = {}
        super(NsxHousekeeper, self).__init__(
            hk_ns, hk_jobs, invoke_on_load=True,
            invoke_args=(self.global_readonly, self.readonly_jobs))

        LOG.info("Loaded housekeeping job names: %s", self.names())
        for job in self:
            if job.obj.get_name() in hk_jobs:
                self.jobs[job.obj.get_name()] = job.obj

    def get(self, job_name):
        if job_name == ALL_DUMMY_JOB_NAME:
            return {'name': job_name,
                    'description': ALL_DUMMY_JOB['description'],
                    'enabled': job_name in self.jobs,
                    'error_count': self.results.get(
                        job_name, {}).get('error_count', 0),
                    'fixed_count': self.results.get(
                        job_name, {}).get('fixed_count', 0),
                    'error_info': self.results.get(
                        job_name, {}).get('error_info', '')}

        for job in self:
            name = job.obj.get_name()
            if job_name == name:
                return {'name': job_name,
                        'description': job.obj.get_description(),
                        'enabled': job_name in self.jobs,
                        'error_count': self.results.get(
                            job_name, {}).get('error_count', 0),
                        'fixed_count': self.results.get(
                            job_name, {}).get('fixed_count', 0),
                        'error_info': self.results.get(
                            job_name, {}).get('error_info', '')}

        raise n_exc.ObjectNotFound(id=job_name)

    def list(self):
        results = [{'name': ALL_DUMMY_JOB_NAME,
                    'description': ALL_DUMMY_JOB['description'],
                    'enabled': ALL_DUMMY_JOB_NAME in self.jobs,
                    'error_count': self.results.get(
                        ALL_DUMMY_JOB_NAME, {}).get('error_count', 0),
                    'fixed_count': self.results.get(
                        ALL_DUMMY_JOB_NAME, {}).get('fixed_count', 0),
                    'error_info': self.results.get(
                        ALL_DUMMY_JOB_NAME, {}).get('error_info', '')}]

        for job in self:
            job_name = job.obj.get_name()
            results.append({'name': job_name,
                            'description': job.obj.get_description(),
                            'enabled': job_name in self.jobs,
                            'error_count': self.results.get(
                                job_name, {}).get('error_count', 0),
                            'fixed_count': self.results.get(
                                job_name, {}).get('fixed_count', 0),
                            'error_info': self.results.get(
                                job_name, {}).get('error_info', '')})

        return results

    def readwrite_allowed(self, job_name):
        # Check if a job can run in readwrite mode
        if self.global_readonly:
            return False

        non_readonly_jobs = set(self.jobs.keys()) - set(self.readonly_jobs)
        if job_name == ALL_DUMMY_JOB_NAME:
            # 'all' readwrite is allowed if it has non readonly jobs
            if non_readonly_jobs:
                return True
            return False
        else:
            # specific job is allowed if it is not in the readonly list
            if job_name in self.readonly_jobs:
                return False
            return True

    def run(self, context, job_name, readonly=False):
        self.results = {}
        if context.is_admin:
            if self.email_notifier:
                self.email_notifier.start('Cloud Housekeeper Execution Report')

            with locking.LockManager.get_lock('nsx-housekeeper'):
                error_count = 0
                fixed_count = 0
                error_info = ''
                if job_name == ALL_DUMMY_JOB_NAME:
                    if (not readonly and
                        not self.readwrite_allowed(ALL_DUMMY_JOB_NAME)):
                        raise n_exc.ObjectNotFound(id=ALL_DUMMY_JOB_NAME)
                    for job in self.jobs.values():
                        if (not readonly and
                            not self.readwrite_allowed(job.get_name())):
                            # skip this job as it is readonly
                            continue
                        result = job.run(context, readonly=readonly)
                        if result:
                            if self.email_notifier and result['error_count']:
                                self._add_job_text_to_notifier(job, result)
                            error_count += result['error_count']
                            fixed_count += result['fixed_count']
                            error_info += result['error_info'] + "\n"
                    self.results[job_name] = {
                        'error_count': error_count,
                        'fixed_count': fixed_count,
                        'error_info': error_info
                    }

                else:
                    job = self.jobs.get(job_name)
                    if job:
                        if (not readonly and
                            not self.readwrite_allowed(job_name)):
                            raise n_exc.ObjectNotFound(id=job_name)
                        result = job.run(context, readonly=readonly)
                        if result:
                            error_count = result['error_count']
                            if self.email_notifier:
                                self._add_job_text_to_notifier(job, result)
                            self.results[job.get_name()] = result
                    else:
                        raise n_exc.ObjectNotFound(id=job_name)

                if self.email_notifier and error_count:
                    self.email_notifier.send()
        else:
            raise n_exc.AdminRequired()

    def _add_job_text_to_notifier(self, job, result):
        self.email_notifier.add_text("<b>%s:</b>", job.get_name())
        self.email_notifier.add_text(
            '%d errors found, %d fixed\n%s\n\n',
            result['error_count'],
            result['fixed_count'],
            result['error_info'])


class HousekeeperEmailNotifier(object):
    def __init__(self):
        self.msg = None
        self.html = None
        self.has_text = False

    def start(self, subject):
        self.msg = MIMEMultipart('alternative')
        self.msg['Subject'] = subject
        self.msg['From'] = cfg.CONF.smtp_from_addr
        self.msg['To'] = ', '.join(cfg.CONF.snmp_to_list)
        self.html = '<html><div>'
        self.has_text = False

    def add_text(self, fmt, *args):
        self.has_text = True
        text = fmt % args
        LOG.debug("Housekeeper emailer adding text %s", text)
        self.html += text.replace("\n", "<br>") + "<br>\n"

    def send(self):
        if self.has_text:
            self.html += "</div></html>"
            part1 = MIMEText(self.html, 'html')
            self.msg.attach(part1)

            s = smtplib.SMTP(cfg.CONF.smtp_gateway)

            s.sendmail(cfg.CONF.smtp_from_addr,
                       cfg.CONF.snmp_to_list,
                       self.msg.as_string())
            s.quit()

        self.msg = None
        self.html = None



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron.tests import base
from neutron_lib.plugins import constants
from oslo_utils import uuidutils

from vmware_nsx.plugins.common.housekeeper import base_job
from vmware_nsx.plugins.nsx_v3.housekeeper import orphaned_logical_router

DUMMY_ROUTER = {
    "resource_type": "LogicalRouter",
    "id": uuidutils.generate_uuid(),
    "display_name": "test",
    "tags": [{
        "scope": "os-neutron-router-id",
        "tag": uuidutils.generate_uuid()
    }, {
        "scope": "os-project-id",
        "tag": uuidutils.generate_uuid()
    }, {
        "scope": "os-project-name",
        "tag": "admin"
    }, {
        "scope": "os-api-version",
        "tag": "13.0.0.0b3.dev90"
    }],
    "edge_cluster_id": uuidutils.generate_uuid(),
    "router_type": "TIER1"}


class OrphanedLogicalRouterTestCaseReadOnly(base.BaseTestCase):

    def setUp(self):
        def get_plugin_mock(alias=constants.CORE):
            if alias in (constants.CORE, constants.L3):
                return self.plugin

        super(OrphanedLogicalRouterTestCaseReadOnly, self).setUp()
        self.plugin = mock.Mock()
        self.plugin.nsxlib = mock.Mock()
        self.context = mock.Mock()
        self.context.session = mock.Mock()
        mock.patch('neutron_lib.plugins.directory.get_plugin',
                   side_effect=get_plugin_mock).start()
        self.log = mock.Mock()
        base_job.LOG = self.log
        self.job = orphaned_logical_router.OrphanedLogicalRouterJob(True, [])

    def run_job(self):
        self.job.run(self.context, readonly=True)

    def test_clean_run(self):
        with mock.patch.object(self.plugin.nsxlib.logical_router, 'list',
                               return_value={'results': []}):
            self.run_job()
            self.log.warning.assert_not_called()

    def test_with_orphaned_ls(self):
        with mock.patch.object(self.plugin.nsxlib.logical_router, 'list',
                               return_value={'results': [DUMMY_ROUTER]}),\
            mock.patch("vmware_nsx.db.db.get_neutron_from_nsx_router_id",
                       return_value=None):
            self.run_job()
            self.log.warning.assert_called()


class OrphanedLogicalRouterTestCaseReadWrite(
    OrphanedLogicalRouterTestCaseReadOnly):

    def run_job(self):
        self.job.run(self.context, readonly=False)



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from neutron.db import l3_db
from neutron.services.flavors import flavors_plugin
from neutron_lib import exceptions as n_exc
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from vmware_nsx._i18n import _
from vmware_nsx.db import db as nsx_db
from vmware_nsx.services.lbaas import lb_const
from vmware_nsxlib.v3 import load_balancer as nsxlib_lb
from vmware_nsxlib.v3 import nsx_constants
from vmware_nsxlib.v3 import utils

LOG = logging.getLogger(__name__)
ADV_RULE_NAME = 'LB external VIP advertisement'
NO_ROUTER_ID = 'NO ROUTER'


@log_helpers.log_method_call
def get_tags(plugin, resource_id, resource_type, project_id, project_name):
    resource = {'project_id': project_id,
                'id': resource_id}
    tags = plugin.nsxlib.build_v3_tags_payload(
        resource, resource_type=resource_type,
        project_name=project_name)
    return tags


@log_helpers.log_method_call
def get_network_from_subnet(context, plugin, subnet_id):
    subnet = plugin.get_subnet(context.elevated(), subnet_id)
    if subnet:
        return plugin.get_network(context.elevated(), subnet['network_id'])


@log_helpers.log_method_call
def get_router_from_network(context, plugin, subnet_id):
    subnet = plugin.get_subnet(context, subnet_id)
    network_id = subnet['network_id']
    port_filters = {'device_owner': [l3_db.DEVICE_OWNER_ROUTER_INTF],
                    'network_id': [network_id]}
    ports = plugin.get_ports(context.elevated(), filters=port_filters)
    if ports:
        router = plugin.get_router(context.elevated(), ports[0]['device_id'])
        if router.get('external_gateway_info'):
            return router['id']


@log_helpers.log_method_call
def get_lb_flavor_size(flavor_plugin, context, flavor_id):
    if not flavor_id:
        return lb_const.DEFAULT_LB_SIZE
    else:
        flavor = flavors_plugin.FlavorsPlugin.get_flavor(
            flavor_plugin, context, flavor_id)
        flavor_size = flavor['name']
        if flavor_size in lb_const.LB_FLAVOR_SIZES:
            return flavor_size.upper()
        else:
            err_msg = (_("Invalid flavor size %(flavor)s, only 'small', "
                         "'medium', or 'large' are supported") %
                       {'flavor': flavor_size})
            raise n_exc.InvalidInput(error_message=err_msg)


@log_helpers.log_method_call
def validate_lb_subnet(context, plugin, subnet_id):
    '''Validate LB subnet before creating loadbalancer on it.

    To create a loadbalancer, the network has to be either an external
    network or private network that connects to a tenant router. The
    tenant router needs to connect to gateway. It will throw
    exception if the network doesn't meet this requirement.

    :param context: context
    :param plugin: core plugin
    :param subnet_id: loadbalancer's subnet id
    :return: True if subnet meet requirement, otherwise return False
    '''
    network = get_network_from_subnet(context, plugin, subnet_id)
    valid_router = get_router_from_network(
        context, plugin, subnet_id)
    if network.get('router:external') or valid_router:
        return True
    else:
        return False


@log_helpers.log_method_call
def validate_lb_member_subnet(context, plugin, subnet_id, lb):
    '''Validate LB member subnet before creating a member.

    The member subnet should belong to an external network or be connected
    to the same T1 router as the Lb vip.
    It will throw exception if the subnet doesn't meet this requirement.

    :param context: context
    :param plugin: core plugin
    :param subnet_id: loadbalancer's subnet id
    :return: True if subnet meet requirement, otherwise return False
    '''
    network = get_network_from_subnet(context, plugin, subnet_id)
    if network.get('router:external'):
        return True
    member_router_id = get_router_from_network(
        context, plugin, subnet_id)
    lb_router_id = get_router_from_network(
        context, plugin, lb['vip_subnet_id'])
    if lb_router_id:
        # Lb on non-external network. member must be on the same router
        if lb_router_id == member_router_id:
            return True
        else:
            return False
    else:
        # LB on external network. member subnet must have a router
        if member_router_id:
            return True
        else:
            return False


def get_rule_match_conditions(policy):
    match_conditions = []
    # values in rule have already been validated in LBaaS API,
    # we won't need to valid anymore in driver, and just get
    # the LB rule mapping from the dict.
    for rule in policy['rules']:
        match_type = lb_const.LB_RULE_MATCH_TYPE[rule['compare_type']]
        if rule['type'] == lb_const.L7_RULE_TYPE_COOKIE:
            header_value = rule['key'] + '=' + rule['value']
            match_conditions.append(
                {'type': 'LbHttpRequestHeaderCondition',
                 'match_type': match_type,
                 'header_name': 'Cookie',
                 'header_value': header_value})
        elif rule['type'] == lb_const.L7_RULE_TYPE_FILE_TYPE:
            match_conditions.append(
                {'type': 'LbHttpRequestUriCondition',
                 'match_type': match_type,
                 'uri': '*.' + rule['value']})
        elif rule['type'] == lb_const.L7_RULE_TYPE_HEADER:
            match_conditions.append(
                {'type': 'LbHttpRequestHeaderCondition',
                 'match_type': match_type,
                 'header_name': rule['key'],
                 'header_value': rule['value']})
        elif rule['type'] == lb_const.L7_RULE_TYPE_HOST_NAME:
            match_conditions.append(
                {'type': 'LbHttpRequestHeaderCondition',
                 'match_type': match_type,
                 'header_name': 'Host',
                 'header_value': rule['value']})
        elif rule['type'] == lb_const.L7_RULE_TYPE_PATH:
            match_conditions.append(
                {'type': 'LbHttpRequestUriCondition',
                 'match_type': match_type,
                 'uri': rule['value']})
        else:
            msg = (_('l7rule type %(type)s is not supported in LBaaS') %
                   {'type': rule['type']})
            raise n_exc.BadRequest(resource='lbaas-l7rule', msg=msg)
    return match_conditions


def get_rule_actions(context, l7policy):
    lb_id = l7policy['listener']['loadbalancer_id']
    if l7policy['action'] == lb_const.L7_POLICY_ACTION_REDIRECT_TO_POOL:
        pool_binding = nsx_db.get_nsx_lbaas_pool_binding(
            context.session, lb_id, l7policy['redirect_pool_id'])
        if pool_binding:
            lb_pool_id = pool_binding['lb_pool_id']
            actions = [{'type': lb_const.LB_SELECT_POOL_ACTION,
                        'pool_id': lb_pool_id}]
        else:
            msg = _('Failed to get LB pool binding from nsx db')
            raise n_exc.BadRequest(resource='lbaas-l7rule-create',
                                   msg=msg)
    elif l7policy['action'] == lb_const.L7_POLICY_ACTION_REDIRECT_TO_URL:
        actions = [{'type': lb_const.LB_HTTP_REDIRECT_ACTION,
                    'redirect_status': lb_const.LB_HTTP_REDIRECT_STATUS,
                    'redirect_url': l7policy['redirect_url']}]
    elif l7policy['action'] == lb_const.L7_POLICY_ACTION_REJECT:
        actions = [{'type': lb_const.LB_REJECT_ACTION,
                    'reply_status': lb_const.LB_HTTP_REJECT_STATUS}]
    else:
        msg = (_('Invalid l7policy action: %(action)s') %
               {'action': l7policy['action']})
        raise n_exc.BadRequest(resource='lbaas-l7rule-create',
                               msg=msg)
    return actions


def convert_l7policy_to_lb_rule(context, policy):
    return {
        'match_conditions': get_rule_match_conditions(policy),
        'actions': get_rule_actions(context, policy),
        'phase': lb_const.LB_RULE_HTTP_FORWARDING,
        'match_strategy': 'ALL'
    }


def remove_rule_from_policy(rule):
    l7rules = rule['policy']['rules']
    rule['policy']['rules'] = [r for r in l7rules if r['id'] != rule['id']]


def update_rule_in_policy(rule):
    remove_rule_from_policy(rule)
    rule['policy']['rules'].append(rule)


@log_helpers.log_method_call
def update_router_lb_vip_advertisement(context, core_plugin, router,
                                       nsx_router_id):
    # Add a rule to advertise external vips on the router
    external_subnets = core_plugin._find_router_gw_subnets(
        context.elevated(), router)
    external_cidrs = [s['cidr'] for s in external_subnets]
    if external_cidrs:
        adv_rule = {
            'display_name': ADV_RULE_NAME,
            'action': nsx_constants.FW_ACTION_ALLOW,
            'networks': external_cidrs,
            'rule_filter': {'prefix_operator': 'GE',
                            'match_route_types': ['T1_LB_VIP']}}
        core_plugin.nsxlib.logical_router.update_advertisement_rules(
            nsx_router_id, [adv_rule], name_prefix=ADV_RULE_NAME)


@log_helpers.log_method_call
def delete_persistence_profile(nsxlib, persistence_profile_id):
    if persistence_profile_id:
        nsxlib.load_balancer.persistence_profile.delete(persistence_profile_id)


def build_persistence_profile_tags(pool_tags, listener):
    tags = pool_tags[:]
    # With octavia loadbalancer name might not be among data passed
    # down to the driver
    lb_data = listener.get('loadbalancer')
    if lb_data:
        tags.append({
            'scope': lb_const.LB_LB_NAME,
            'tag': lb_data['name'][:utils.MAX_TAG_LEN]})
    tags.append({
            'scope': lb_const.LB_LB_TYPE,
            'tag': listener['loadbalancer_id']})
    tags.append({
            'scope': lb_const.LB_LISTENER_TYPE,
            'tag': listener['id']})
    return tags


def get_pool_tags(context, core_plugin, pool):
    return get_tags(core_plugin, pool['id'],
                    lb_const.LB_POOL_TYPE, pool.get('tenant_id', ''),
                    context.project_name)


def setup_session_persistence(nsxlib, pool, pool_tags,
                              switch_type, listener, vs_data):
    sp = pool.get('session_persistence')
    pers_type = None
    cookie_name = None
    cookie_mode = None
    if not sp:
        LOG.debug("No session persistence info for pool %s", pool['id'])
    elif sp['type'] == lb_const.LB_SESSION_PERSISTENCE_HTTP_COOKIE:
        pers_type = nsxlib_lb.PersistenceProfileTypes.COOKIE
        cookie_name = sp.get('cookie_name')
        if not cookie_name:
            cookie_name = lb_const.SESSION_PERSISTENCE_DEFAULT_COOKIE_NAME
        cookie_mode = "INSERT"
    elif sp['type'] == lb_const.LB_SESSION_PERSISTENCE_APP_COOKIE:
        pers_type = nsxlib_lb.PersistenceProfileTypes.COOKIE
        # In this case cookie name is mandatory
        cookie_name = sp['cookie_name']
        cookie_mode = "REWRITE"
    else:
        pers_type = nsxlib_lb.PersistenceProfileTypes.SOURCE_IP

    if pers_type:
        # There is a profile to create or update
        pp_kwargs = {
            'resource_type': pers_type,
            'display_name': "persistence_%s" % utils.get_name_and_uuid(
                pool['name'] or 'pool', pool['id'], maxlen=235),
            'tags': build_persistence_profile_tags(pool_tags, listener)
        }
        if cookie_name:
            pp_kwargs['cookie_name'] = cookie_name
            pp_kwargs['cookie_mode'] = cookie_mode

    pp_client = nsxlib.load_balancer.persistence_profile
    persistence_profile_id = vs_data.get('persistence_profile_id')
    if persistence_profile_id and not switch_type:
        # NOTE: removal of the persistence profile must be executed
        # after the virtual server has been updated
        if pers_type:
            # Update existing profile
            LOG.debug("Updating persistence profile %(profile_id)s for "
                      "listener %(listener_id)s with pool %(pool_id)s",
                      {'profile_id': persistence_profile_id,
                       'listener_id': listener['id'],
                       'pool_id': pool['id']})
            pp_client.update(persistence_profile_id, **pp_kwargs)
            return persistence_profile_id, None
        else:
            # Prepare removal of persistence profile
            return (None, functools.partial(delete_persistence_profile,
                                            nsxlib, persistence_profile_id))
    elif pers_type:
        # Create persistence profile
        pp_data = pp_client.create(**pp_kwargs)
        LOG.debug("Created persistence profile %(profile_id)s for "
                  "listener %(listener_id)s with pool %(pool_id)s",
                  {'profile_id': pp_data['id'],
                   'listener_id': listener['id'],
                   'pool_id': pool['id']})
        if switch_type:
            # There is also a persistence profile to remove!
            return (pp_data['id'],
                    functools.partial(delete_persistence_profile,
                                      nsxlib, persistence_profile_id))
        return pp_data['id'], None
    return None, None



# Copyright 2019 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_fwaas.db.firewall.v2 import firewall_db_v2
from neutron_fwaas.services.firewall.service_drivers.agents import agents
from neutron_lib import constants as nl_constants


class ApiReplayFirewallPluginDb(firewall_db_v2.FirewallPluginDb):
    """Override FWaaS agent DB actions to use given objects IDs"""
    def create_firewall_rule(self, context, firewall_rule):
        fwr = firewall_rule
        src_port_min, src_port_max = self._get_min_max_ports_from_range(
            fwr['source_port'])
        dst_port_min, dst_port_max = self._get_min_max_ports_from_range(
            fwr['destination_port'])
        with context.session.begin(subtransactions=True):
            fwr_db = firewall_db_v2.FirewallRuleV2(
                # Use given ID for api_replay support
                id=fwr.get('id'),
                tenant_id=fwr['tenant_id'],
                name=fwr['name'],
                description=fwr['description'],
                protocol=fwr['protocol'],
                ip_version=fwr['ip_version'],
                source_ip_address=fwr['source_ip_address'],
                destination_ip_address=fwr['destination_ip_address'],
                source_port_range_min=src_port_min,
                source_port_range_max=src_port_max,
                destination_port_range_min=dst_port_min,
                destination_port_range_max=dst_port_max,
                action=fwr['action'],
                enabled=fwr['enabled'],
                shared=fwr['shared'])
            context.session.add(fwr_db)
        return self._make_firewall_rule_dict(fwr_db)

    def create_firewall_policy(self, context, firewall_policy):
        """This method is manipulated to allow the creation of additional
        default firewall policy, and do not automatically ensure one exists
        """
        fwp = firewall_policy
        with context.session.begin(subtransactions=True):
            # Use given ID for api_replay support
            fwp_db = firewall_db_v2.FirewallPolicy(
                id=fwp.get('id'),
                tenant_id=fwp['tenant_id'],
                name=fwp['name'],
                description=fwp['description'],
                audited=fwp['audited'],
                shared=fwp['shared'])
            context.session.add(fwp_db)
            self._set_rules_for_policy(context, fwp_db, fwp)
        return self._make_firewall_policy_dict(fwp_db)

    def create_firewall_group(self, context, firewall_group,
                              default_fwg=False):
        """This method is manipulated to allow the creation of additional
        default firewall group, and do not automatically ensure one exists
        """
        fwg = firewall_group
        tenant_id = fwg['tenant_id']
        if firewall_group.get('status') is None:
            fwg['status'] = nl_constants.CREATED

        with context.session.begin(subtransactions=True):
            # Use given ID for api_replay support
            fwg_db = firewall_db_v2.FirewallGroup(
                id=fwg.get('id'),
                tenant_id=tenant_id,
                name=fwg['name'],
                description=fwg['description'],
                status=fwg['status'],
                ingress_firewall_policy_id=fwg['ingress_firewall_policy_id'],
                egress_firewall_policy_id=fwg['egress_firewall_policy_id'],
                admin_state_up=fwg['admin_state_up'],
                shared=fwg['shared'])
            context.session.add(fwg_db)
            self._set_ports_for_firewall_group(context, fwg_db, fwg)
        return self._make_firewall_group_dict(fwg_db)


class ApiReplayFirewallAgentDriver(agents.FirewallAgentDriver):
    """FWaaS V2 agent driver for api-replay allowing POST with id."""
    def __init__(self, *args, **kwargs):
        super(ApiReplayFirewallAgentDriver, self).__init__(*args, **kwargs)
        self.firewall_db = ApiReplayFirewallPluginDb()



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

from neutron_lib.plugins import directory

from vmware_nsx.db import nsxv_models
from vmware_nsx.plugins.nsx_v.vshield import edge_firewall_driver
from vmware_nsx.plugins.nsx_v.vshield import edge_utils
from vmware_nsx.services.fwaas.nsx_v import edge_fwaas_driver_v2
from vmware_nsx.services.fwaas.nsx_v import fwaas_callbacks_v2
from vmware_nsx.tests.unit.nsx_v import test_plugin as test_v_plugin

FAKE_FW_ID = 'fake_fw_uuid'
FAKE_ROUTER_ID = 'fake_rtr_uuid'
FAKE_PORT_ID = 'fake_port_uuid'
FAKE_NET_ID = 'fake_net_uuid'
FAKE_DB_OBJ = nsxv_models.NsxvEdgeVnicBinding(vnic_index='1')


class NsxvFwaasTestCase(test_v_plugin.NsxVPluginV2TestCase):
    def setUp(self):
        super(NsxvFwaasTestCase, self).setUp()
        self.firewall = edge_fwaas_driver_v2.EdgeFwaasVDriverV2()

        self.plugin = directory.get_plugin()
        self.plugin.fwaas_callbacks = fwaas_callbacks_v2.\
            NsxvFwaasCallbacksV2(False)
        self.plugin.fwaas_callbacks.fwaas_enabled = True
        self.plugin.fwaas_callbacks.fwaas_driver = self.firewall
        self.plugin.fwaas_callbacks.internal_driver = self.firewall
        self.plugin.init_is_complete = True
        self.plugin.metadata_proxy_handler = None

        # Start some mocks
        self.router = {'id': FAKE_ROUTER_ID,
                       'external_gateway_info': {'network_id': 'external'},
                       'nsx_attributes': {'distributed': False,
                                          'router_type': 'exclusive'}}
        self.distributed_router = {'id': FAKE_ROUTER_ID,
                       'external_gateway_info': {'network_id': 'external'},
                       'nsx_attributes': {'distributed': True,
                                          'router_type': 'exclusive'}}
        mock.patch.object(self.plugin, '_get_router',
                          return_value=self.router).start()
        mock.patch.object(self.plugin, 'get_router',
                          return_value=self.router).start()
        self.port = {'id': FAKE_PORT_ID, 'network_id': FAKE_NET_ID}
        mock.patch.object(self.plugin, '_get_router_interfaces',
                          return_value=[self.port]).start()
        mock.patch.object(self.plugin, 'get_port',
                          return_value=self.port).start()
        mock.patch.object(self.plugin, '_get_subnet_fw_rules',
                          return_value=[]).start()
        mock.patch.object(self.plugin, '_get_dnat_fw_rule',
                          return_value=[]).start()
        mock.patch.object(self.plugin, '_get_allocation_pools_fw_rule',
                          return_value=[]).start()
        mock.patch.object(self.plugin, '_get_nosnat_subnets_fw_rules',
                          return_value=[]).start()

    def _fake_rules_v4(self, is_ingress=True, is_conflict=False,
                       cidr='10.24.4.0/24'):
        rule1 = {'enabled': True,
                 'action': 'allow',
                 'ip_version': 4,
                 'protocol': 'tcp',
                 'destination_port': '80',
                 'id': 'fake-fw-rule1',
                 'description': 'first rule',
                 'position': '0'}
        rule2 = {'enabled': True,
                 'action': 'reject',
                 'ip_version': 4,
                 'protocol': 'tcp',
                 'destination_port': '22:24',
                 'source_port': '1:65535',
                 'id': 'fake-fw-rule2',
                 'position': '1'}
        rule3 = {'enabled': True,
                 'action': 'deny',
                 'ip_version': 4,
                 'protocol': 'icmp',
                 'id': 'fake-fw-rule3',
                 'position': '2'}
        rule4 = {'enabled': True,
                 'action': 'deny',
                 'ip_version': 4,
                 'id': 'fake-fw-rule4',
                 'position': '3'}
        if is_ingress:
            if not is_conflict:
                rule1['source_ip_address'] = cidr
            else:
                rule1['destination_ip_address'] = cidr
        else:
            if not is_conflict:
                rule1['destination_ip_address'] = cidr
            else:
                rule1['source_ip_address'] = cidr

        return [rule1, rule2, rule3, rule4]

    def _fake_translated_rules(self, rules_list,
                               nsx_port_id,
                               is_ingress=True,
                               logged=False):
        translated_rules = copy.copy(rules_list)
        for rule in translated_rules:
            if logged:
                rule['logged'] = True
            if is_ingress:
                if (not rule.get('destination_ip_address') or
                    rule['destination_ip_address'].startswith('0.0.0.0')):
                    if nsx_port_id:
                        rule['destination_vnic_groups'] = [nsx_port_id]
            else:
                if (not rule.get('source_ip_address') or
                    rule['source_ip_address'].startswith('0.0.0.0')):
                    if nsx_port_id:
                        rule['source_vnic_groups'] = [nsx_port_id]
            if rule.get('destination_ip_address'):
                if rule['destination_ip_address'].startswith('0.0.0.0'):
                    del rule['destination_ip_address']
                else:
                    rule['destination_ip_address'] = [
                        rule['destination_ip_address']]
            if rule.get('source_ip_address'):
                if rule['source_ip_address'].startswith('0.0.0.0'):
                    del rule['source_ip_address']
                else:
                    rule['source_ip_address'] = [
                        rule['source_ip_address']]
            rule['name'] = (fwaas_callbacks_v2.RULE_NAME_PREFIX +
                            (rule.get('name') or rule['id']))[:30]
            if rule.get('id'):
                if is_ingress:
                    rule['id'] = ('ingress-%s' % rule['id'])[:36]
                else:
                    rule['id'] = ('egress-%s' % rule['id'])[:36]

        return translated_rules

    def _fake_empty_firewall_group(self):
        fw_inst = {'id': FAKE_FW_ID,
                   'admin_state_up': True,
                   'tenant_id': 'tenant-uuid',
                   'ingress_rule_list': [],
                   'egress_rule_list': []}
        return fw_inst

    def _fake_firewall_group(self, rule_list, is_ingress=True,
                             admin_state_up=True):
        _rule_list = copy.deepcopy(rule_list)
        for rule in _rule_list:
            rule['position'] = str(_rule_list.index(rule))
        fw_inst = {'id': FAKE_FW_ID,
                   'admin_state_up': admin_state_up,
                   'tenant_id': 'tenant-uuid',
                   'ingress_rule_list': [],
                   'egress_rule_list': []}
        if is_ingress:
            fw_inst['ingress_rule_list'] = _rule_list
        else:
            fw_inst['egress_rule_list'] = _rule_list
        return fw_inst

    def _fake_firewall_group_with_admin_down(self, rule_list,
                                             is_ingress=True):
        return self._fake_firewall_group(
            rule_list, is_ingress=is_ingress, admin_state_up=False)

    def _fake_apply_list_template(self, router):
        router_inst = router
        router_info_inst = mock.Mock()
        router_info_inst.router = router_inst
        router_info_inst.router_id = FAKE_ROUTER_ID
        apply_list = [(router_info_inst, FAKE_PORT_ID)]
        return apply_list

    def _fake_apply_list(self):
        return self._fake_apply_list_template(self.router)

    def _fake_distributed_apply_list(self):
        return self._fake_apply_list_template(self.distributed_router)

    def test_create_firewall_no_rules(self):
        apply_list = self._fake_apply_list()
        firewall = self._fake_empty_firewall_group()
        with mock.patch.object(self.plugin.fwaas_callbacks, 'get_port_fwg',
                              return_value=firewall),\
            mock.patch.object(self.plugin.fwaas_callbacks,
                              '_get_port_firewall_group_id',
                              return_value=FAKE_FW_ID),\
            mock.patch.object(self.plugin.fwaas_callbacks,
                              '_get_fw_group_from_plugin',
                              return_value=firewall),\
            mock.patch("vmware_nsx.db.nsxv_db.get_edge_vnic_binding",
                       return_value=FAKE_DB_OBJ),\
            mock.patch.object(edge_utils, "update_firewall") as update_fw,\
            mock.patch.object(edge_utils, 'get_router_edge_id',
                              return_value='edge-1'):
            self.firewall.create_firewall_group('nsx', apply_list, firewall)
            # expecting 2 block rules for the logical port (egress & ingress)
            # and last default allow all rule
            expected_rules = [
                {'name': "Block port ingress",
                 'action': edge_firewall_driver.FWAAS_DENY,
                 'destination_vnic_groups': ['vnic-index-1'],
                 'logged': False},
                {'name': "Block port egress",
                 'action': edge_firewall_driver.FWAAS_DENY,
                 'source_vnic_groups': ['vnic-index-1'],
                 'logged': False}]
            update_fw.assert_called_once_with(
                self.plugin.nsx_v, mock.ANY, FAKE_ROUTER_ID,
                {'firewall_rule_list': expected_rules})

    def _setup_firewall_with_rules(self, func, is_ingress=True,
                                   is_conflict=False, cidr='10.24.4.0/24'):
        apply_list = self._fake_apply_list()
        rule_list = self._fake_rules_v4(is_ingress=is_ingress,
                                        is_conflict=is_conflict, cidr=cidr)
        firewall = self._fake_firewall_group(rule_list, is_ingress=is_ingress)
        with mock.patch.object(self.plugin.fwaas_callbacks, 'get_port_fwg',
                              return_value=firewall),\
            mock.patch.object(self.plugin.fwaas_callbacks,
                              '_get_port_firewall_group_id',
                              return_value=FAKE_FW_ID),\
            mock.patch.object(self.plugin.fwaas_callbacks,
                              '_get_fw_group_from_plugin',
                              return_value=firewall),\
            mock.patch("vmware_nsx.db.nsxv_db.get_edge_vnic_binding",
                       return_value=FAKE_DB_OBJ),\
            mock.patch.object(edge_utils, "update_firewall") as update_fw,\
            mock.patch.object(edge_utils, 'get_router_edge_id',
                              return_value='edge-1'):
            func('nsx', apply_list, firewall)
            expected_rules = self._fake_translated_rules(
                rule_list,
                'vnic-index-1', is_ingress=is_ingress) + [
                {'name': "Block port ingress",
                 'action': edge_firewall_driver.FWAAS_DENY,
                 'destination_vnic_groups': ['vnic-index-1'],
                 'logged': False},
                {'name': "Block port egress",
                 'action': edge_firewall_driver.FWAAS_DENY,
                 'source_vnic_groups': ['vnic-index-1'],
                 'logged': False}]

            update_fw.assert_called_once_with(
                self.plugin.nsx_v, mock.ANY, FAKE_ROUTER_ID,
                {'firewall_rule_list': expected_rules})

    def test_create_firewall_with_ingress_rules(self):
        self._setup_firewall_with_rules(self.firewall.create_firewall_group)

    def test_update_firewall_with_ingress_rules(self):
        self._setup_firewall_with_rules(self.firewall.update_firewall_group)

    def test_create_firewall_with_egress_rules(self):
        self._setup_firewall_with_rules(self.firewall.create_firewall_group,
                                        is_ingress=False)

    def test_create_firewall_with_illegal_cidr(self):
        self._setup_firewall_with_rules(self.firewall.create_firewall_group,
                                        cidr='0.0.0.0/24')

    def test_update_firewall_with_egress_rules(self):
        self._setup_firewall_with_rules(self.firewall.update_firewall_group,
                                        is_ingress=False)

    def test_update_firewall_with_egress_conflicting_rules(self):
        self._setup_firewall_with_rules(self.firewall.update_firewall_group,
                                        is_ingress=False, is_conflict=True)

    def test_update_firewall_with_ingress_conflicting_rules(self):
        self._setup_firewall_with_rules(self.firewall.update_firewall_group,
                                        is_ingress=True, is_conflict=True)

    def test_delete_firewall(self):
        apply_list = self._fake_apply_list()
        firewall = self._fake_empty_firewall_group()
        with mock.patch.object(self.plugin.fwaas_callbacks, 'get_port_fwg',
                              return_value=None),\
            mock.patch("vmware_nsx.db.db.get_nsx_switch_and_port_id",
                       return_value=('vnic-index-1', 0)),\
            mock.patch.object(edge_utils, "update_firewall") as update_fw,\
            mock.patch.object(edge_utils, 'get_router_edge_id',
                              return_value='edge-1'):
            self.firewall.delete_firewall_group('nsx', apply_list, firewall)
            update_fw.assert_called_once_with(
                self.plugin.nsx_v, mock.ANY, FAKE_ROUTER_ID,
                {'firewall_rule_list': []})

    def test_create_firewall_with_admin_down(self):
        apply_list = self._fake_apply_list()
        rule_list = self._fake_rules_v4()
        firewall = self._fake_firewall_group_with_admin_down(rule_list)
        with mock.patch.object(edge_utils, "update_firewall") as update_fw,\
            mock.patch.object(edge_utils, 'get_router_edge_id',
                              return_value='edge-1'):
            self.firewall.create_firewall_group('nsx', apply_list, firewall)
            update_fw.assert_called_once_with(
                self.plugin.nsx_v, mock.ANY, FAKE_ROUTER_ID,
                {'firewall_rule_list': []})

    def _setup_dist_router_firewall_with_rules(self, func, is_ingress=True,
                                               is_conflict=False,
                                               cidr='10.24.4.0/24'):
        apply_list = self._fake_distributed_apply_list()
        rule_list = self._fake_rules_v4(is_ingress=is_ingress,
                                        is_conflict=is_conflict, cidr=cidr)
        firewall = self._fake_firewall_group(rule_list, is_ingress=is_ingress)
        with mock.patch.object(self.plugin.fwaas_callbacks, 'get_port_fwg',
                              return_value=firewall),\
            mock.patch.object(self.plugin.fwaas_callbacks,
                              '_get_port_firewall_group_id',
                              return_value=FAKE_FW_ID),\
            mock.patch.object(self.plugin.fwaas_callbacks,
                              '_get_fw_group_from_plugin',
                              return_value=firewall),\
            mock.patch.object(edge_utils, "update_firewall") as update_fw,\
            mock.patch.object(edge_utils, 'get_router_edge_id',
                              return_value='edge-1'),\
            mock.patch.object(self.plugin.edge_manager, 'get_plr_by_tlr_id',
                              return_value=FAKE_ROUTER_ID),\
            mock.patch.object(self.plugin, '_get_router',
                              return_value=self.distributed_router),\
            mock.patch.object(self.plugin, 'get_router',
                              return_value=self.distributed_router):
            func('nsx', apply_list, firewall)
            expected_rules = self._fake_translated_rules(
                rule_list, None, is_ingress=is_ingress) + [
                {'name': "Block port ingress",
                 'action': edge_firewall_driver.FWAAS_DENY,
                 'logged': False},
                {'name': "Block port egress",
                 'action': edge_firewall_driver.FWAAS_DENY,
                 'logged': False}]

            update_fw.assert_called_once_with(
                self.plugin.nsx_v, mock.ANY, FAKE_ROUTER_ID,
                {'firewall_rule_list': expected_rules})

    def test_create_dist_router_firewall_with_ingress_rules(self):
        self._setup_dist_router_firewall_with_rules(
            self.firewall.create_firewall_group)

    def test_update_dist_router_firewall_with_ingress_rules(self):
        self._setup_dist_router_firewall_with_rules(
            self.firewall.update_firewall_group)

    def test_create_dist_router_firewall_with_egress_rules(self):
        self._setup_dist_router_firewall_with_rules(
            self.firewall.create_firewall_group,
            is_ingress=False)

    def test_create_dist_router_firewall_with_illegal_cidr(self):
        self._setup_dist_router_firewall_with_rules(
            self.firewall.create_firewall_group,
            cidr='0.0.0.0/24')

    def test_update_dist_router_firewall_with_egress_rules(self):
        self._setup_dist_router_firewall_with_rules(
            self.firewall.update_firewall_group,
            is_ingress=False)

    def test_update_dist_router_firewall_with_egress_conflicting_rules(self):
        self._setup_dist_router_firewall_with_rules(
            self.firewall.update_firewall_group,
            is_ingress=False, is_conflict=True)

    def test_update_dist_router_firewall_with_ingress_conflicting_rules(self):
        self._setup_dist_router_firewall_with_rules(
            self.firewall.update_firewall_group,
            is_ingress=True, is_conflict=True)



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_consts
from neutron_lib import exceptions as n_exc
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas import lb_helper
from vmware_nsx.services.lbaas import lb_translators
from vmware_nsx.services.lbaas.nsx_p.implementation import healthmonitor_mgr
from vmware_nsx.services.lbaas.nsx_p.implementation import l7policy_mgr
from vmware_nsx.services.lbaas.nsx_p.implementation import l7rule_mgr
from vmware_nsx.services.lbaas.nsx_p.implementation import listener_mgr
from vmware_nsx.services.lbaas.nsx_p.implementation import loadbalancer_mgr
from vmware_nsx.services.lbaas.nsx_p.implementation import member_mgr
from vmware_nsx.services.lbaas.nsx_p.implementation import pool_mgr
from vmware_nsx.services.lbaas.octavia import constants as oct_const

LOG = logging.getLogger(__name__)


class NotImplementedManager(object):
    """Helper class to make any subclass of LoadBalancerBaseDriver explode if
    it is missing any of the required object managers.
    """

    def create(self, context, obj):
        raise NotImplementedError()

    def update(self, context, old_obj, obj):
        raise NotImplementedError()

    def delete(self, context, obj):
        raise NotImplementedError()


class EdgeLoadbalancerDriverV2(base_mgr.LoadbalancerBaseManager):
    @log_helpers.log_method_call
    def __init__(self):
        super(EdgeLoadbalancerDriverV2, self).__init__()

        # Init all LBaaS objects
        # Note(asarfaty): self.lbv2_driver is not yet defined at init time
        # so lambda is used to retrieve it later.
        self.loadbalancer = lb_helper.LBaaSNSXObjectManagerWrapper(
            "loadbalancer",
            loadbalancer_mgr.EdgeLoadBalancerManagerFromDict(),
            lb_translators.lb_loadbalancer_obj_to_dict,
            lambda: self.lbv2_driver.load_balancer)

        self.listener = lb_helper.LBaaSNSXObjectManagerWrapper(
            "listener",
            listener_mgr.EdgeListenerManagerFromDict(),
            lb_translators.lb_listener_obj_to_dict,
            lambda: self.lbv2_driver.listener)

        self.pool = lb_helper.LBaaSNSXObjectManagerWrapper(
            "pool",
            pool_mgr.EdgePoolManagerFromDict(),
            lb_translators.lb_pool_obj_to_dict,
            lambda: self.lbv2_driver.pool)

        self.member = lb_helper.LBaaSNSXObjectManagerWrapper(
            "member",
            member_mgr.EdgeMemberManagerFromDict(),
            lb_translators.lb_member_obj_to_dict,
            lambda: self.lbv2_driver.member)

        self.healthmonitor = lb_helper.LBaaSNSXObjectManagerWrapper(
            "healthmonitor",
            healthmonitor_mgr.EdgeHealthMonitorManagerFromDict(),
            lb_translators.lb_hm_obj_to_dict,
            lambda: self.lbv2_driver.health_monitor)

        self.l7policy = lb_helper.LBaaSNSXObjectManagerWrapper(
            "l7policy",
            l7policy_mgr.EdgeL7PolicyManagerFromDict(),
            lb_translators.lb_l7policy_obj_to_dict,
            lambda: self.lbv2_driver.l7policy)

        self.l7rule = lb_helper.LBaaSNSXObjectManagerWrapper(
            "l7rule",
            l7rule_mgr.EdgeL7RuleManagerFromDict(),
            lb_translators.lb_l7rule_obj_to_dict,
            lambda: self.lbv2_driver.l7rule)

        self._subscribe_router_delete_callback()

    def _subscribe_router_delete_callback(self):
        # Check if there is any LB attachment for the NSX router.
        # This callback is subscribed here to prevent router/GW/interface
        # deletion if it still has LB service attached to it.

        #Note(asarfaty): Those callbacks are used by Octavia as well even
        # though they are bound only here
        registry.subscribe(self._check_lb_service_on_router,
                           resources.ROUTER, events.BEFORE_DELETE)
        registry.subscribe(self._check_lb_service_on_router,
                           resources.ROUTER_GATEWAY, events.BEFORE_DELETE)
        registry.subscribe(self._check_lb_service_on_router_interface,
                           resources.ROUTER_INTERFACE, events.BEFORE_DELETE)

    def _unsubscribe_router_delete_callback(self):
        registry.unsubscribe(self._check_lb_service_on_router,
                             resources.ROUTER, events.BEFORE_DELETE)
        registry.unsubscribe(self._check_lb_service_on_router,
                             resources.ROUTER_GATEWAY, events.BEFORE_DELETE)
        registry.unsubscribe(self._check_lb_service_on_router_interface,
                             resources.ROUTER_INTERFACE, events.BEFORE_DELETE)

    def _get_lb_ports(self, context, subnet_ids):
        dev_owner_v2 = n_consts.DEVICE_OWNER_LOADBALANCERV2
        dev_owner_oct = oct_const.DEVICE_OWNER_OCTAVIA
        filters = {'device_owner': [dev_owner_v2, dev_owner_oct],
                   'fixed_ips': {'subnet_id': subnet_ids}}
        return self.loadbalancer.core_plugin.get_ports(
            context, filters=filters)

    def _check_lb_service_on_router(self, resource, event, trigger,
                                    payload=None):
        """Prevent removing a router GW or deleting a router used by LB"""
        router_id = payload.resource_id
        core_plugin = self.loadbalancer.core_plugin
        # get the real core plugin
        if core_plugin.service_router_has_loadbalancers(router_id):
            msg = _('Cannot delete a %s as it still has lb service '
                    'attachment') % resource
            raise n_exc.BadRequest(resource='lbaas-lb', msg=msg)

    def _check_lb_service_on_router_interface(
            self, resource, event, trigger, payload=None):
        # Prevent removing the interface of an LB subnet from a router
        router_id = payload.resource_id
        subnet_id = payload.metadata.get('subnet_id')
        if not router_id or not subnet_id:
            return

        # get LB ports and check if any loadbalancer is using this subnet
        if self._get_lb_ports(payload.context.elevated(), [subnet_id]):
            msg = _('Cannot delete a router interface as it used by a '
                    'loadbalancer')
            raise n_exc.BadRequest(resource='lbaas-lb', msg=msg)



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.plugins.common.housekeeper import base_job
from vmware_nsx.plugins.nsx_v3 import utils as v3_utils

LOG = log.getLogger(__name__)


class OrphanedFirewallSectionJob(base_job.BaseJob):

    def __init__(self, global_readonly, readonly_jobs):
        super(OrphanedFirewallSectionJob, self).__init__(
            global_readonly, readonly_jobs)

    def get_project_plugin(self, plugin):
        return plugin.get_plugin_by_type(projectpluginmap.NsxPlugins.NSX_T)

    def get_name(self):
        return 'orphaned_firewall_section'

    def get_description(self):
        return 'Detect orphaned firewall sections'

    def run(self, context, readonly=False):
        super(OrphanedFirewallSectionJob, self).run(context)

        # get all orphaned firewall sections
        orphaned_sections = v3_utils.get_orphaned_firewall_sections(
            context, self.plugin.nsxlib)

        info = ""
        if not orphaned_sections:
            msg = 'No orphaned firewall sections detected.'
            info = base_job.housekeeper_info(info, msg)
            return {'error_count': 0, 'fixed_count': 0, 'error_info': info}

        msg = ("Found %(len)s orphaned firewall section%(plural)s:" %
               {'len': len(orphaned_sections),
                'plural': 's' if len(orphaned_sections) > 1 else ''})
        info = base_job.housekeeper_warning(info, msg)
        fixed_count = 0
        for section in orphaned_sections:
            msg = ("Firewall section %(name)s [id: %(id)s] "
                   "neutron security group: %(sg)s" %
                   {'name': section['display_name'],
                    'id': section['id'],
                    'sg': section['neutron_sg_id'] if section['neutron_sg_id']
                    else 'Unknown'})
            if not readonly:
                try:
                    self.plugin.nsxlib.firewall_section.delete(section['id'])
                except Exception as e:
                    msg = "%s failed to be removed: %s." % (msg, e)
                else:
                    fixed_count = fixed_count + 1
                    msg = "%s was removed." % msg
            info = base_job.housekeeper_warning(info, msg)

        return {'error_count': len(orphaned_sections),
                'error_info': info,
                'fixed_count': fixed_count}



# Copyright 2016 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""nsxv3_secgroup_local_ip_prefix

Revision ID: 081af0e396d7
Revises: 5ed1ffbc0d2a
Create Date: 2016-03-24 07:11:30.300482

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = '081af0e396d7'
down_revision = '5ed1ffbc0d2a'


def upgrade():
    op.rename_table('nsxv_extended_security_group_rule_properties',
                    'nsx_extended_security_group_rule_properties')
